
/* PICMICRO FORTH COMPILER / INTERPRETER

   there are 2 machines involved
     host running compiler/interpreter
     target running target forth

   2 dictionaries stored here are
     host:    host code (compiler) and target macros (target + language)
     target:  compiled target code

   the host execution model is the same as the scripted target execution model:
   primitives are implemented as native subroutines (call threaded) using the c (host) 
   or hardware (target) return stack.

   on top of this runs the 'script' interpreter: a threaded forth using the primitives/
   call threaded composite words of the underlying forth.

   obviously the code of target and script is never executed
   in the host (well, maybe later).

   why like this?
   * call threaded is fast for target (small macros + fast hardware return stack)
   * target has small data ram size (for threaded words, so code size is an issue)
   * host is implemented in c (primitives are call threaded)
   * no access to native code compiler for host, so some threading technique is better
     and it seems like a good idea to use the same technique as in the target
   * no access to return stack for host

   in the target, the inner interpreter is called whenever a threaded
   word is called (through a stub). in the host the inner interpreter
   will determine if a word is a primitive (direct) or a colon definition
   (indirect), based on the address. as far as interpreters go, that's not
   the best solution but for combining both threaded & native code

*/


/**** HEX NEWLINE ISSUES ********/


/* use this if your host system is a unix 
   i've set this to default because MPLAB simulator
   does not like unix newlines. (linux or cygwin)
*/
#define HEX_NEWLINE "%c%c",13,10

/* use this if previous does not work,
   and explore the dos/unix newline mess
   for mingw, use this option
 */
//#define HEX_NEWLINE "\n"   




#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

#define TRUE -1 /* all one */
#define FALSE 0 /* all zero */

typedef struct
{
    char *name;
    void *address;
} item_t;

struct dictionary_;


typedef void (*prim_t)();    // address of primitive code

// the data cell has the size of the host pointer size. (32 bit)
// the target dicts have a smaller size (8 for ram, 16 for native code)

typedef struct dictionary_
{
    int item_size;   // nb dict items
    item_t *item;    // lookup table
    item_t *item_free;        // next free dictionary item

} dictionary_t;






#define HOST 0    // compiler (host) dictionary
#define TARGET 1  // native code running on target

int trace = 0;

void error_undef(char *name){fprintf(stderr, "ERROR: undefined word %s\n", name);}
void error_file(char *name) {fprintf(stderr, "ERROR: can't open file %s\n", name);}
void error_fd(int fd)       {fprintf(stderr, "ERROR: can't open file descriptor %d\n", fd);}

/* stacks */
typedef struct {
    int size;
    void **data;
    void **top;
} stack_t;

stack_t *stack_new(int size)
{
    stack_t *stack = calloc(1, sizeof(*stack));
    stack->size = size;
    stack->data = calloc(size, sizeof(void*));
    stack->top = stack->data + size;
    return stack;
}

stack_t *ds, *rs, *is, *os;
void word_print_ds();
void word_print_rs();

void *stack_top(stack_t *s){return *(s->top);}
void stack_push(stack_t *s, void *thing){
    if (s->top <= s->data){
	fprintf(stderr, "ERROR: stack overflow. contents: \n");
	if (s == ds) word_print_ds();
	if (s == rs) word_print_rs();
    }
    else {
	s->top--;
	*(s->top) = thing;
    }

}
void *stack_pop(stack_t *s){
    if (s->top >= (s->data + s->size)) fprintf(stderr, "ERROR: stack underflow.\n");
    else {
	void *x = *(s->top);
	s->top++;
	return x;
    }
}
void output_printf(char *fmt, ...);
void stack_print(char *fmt, stack_t *s){
    void **i = s->data + s->size;
    while (--i >= s->top) output_printf(fmt, *i);
    output_printf("\n");
}



dictionary_t *dict[3];
int dict_current = 0;


void push_rs(prim_t *p){stack_push(rs, p);}
void push_ds(int i)    {stack_push(ds, (void *)i);}
void push_is(FILE *f)  {stack_push(is, (void *)f);}
void push_os(FILE *f)  {stack_push(os, (void *)f);}

prim_t* pop_rs() {return (prim_t *)stack_pop(rs);}
int     pop_ds() {return (int)stack_pop(ds);}
FILE*   pop_is() {return (FILE *)stack_pop(is);}
FILE*   pop_os() {return (FILE *)stack_pop(os);}

prim_t* top_rs() {return *((prim_t **)(rs->top));}
int     top_ds() {return *((int *)(ds->top));}
FILE*   top_is() {return *((FILE **)(is->top));}
FILE*   top_os() {return *((FILE **)(os->top));}

/* host threaded code buffer */
#define HOST_SIZE 0x8000
prim_t host_code[HOST_SIZE];
prim_t *host_free = host_code;

/* threading:
   ip points to a dictionary cell containing a prim_t (primitive)
   or a prim_t* (composite) word  */
prim_t *host_ip;

/* target code buffer */
#define TARGET_SIZE 0x4000
unsigned short target_code[TARGET_SIZE]; // 16k WORDS
unsigned short *target_free = target_code;



char word[64]; // last read input word




/* * * * * * * * * * * * * *
*         OUTPUT           *
* * * * * * * * * * * * * */

void drop_os(){
    fclose (pop_os());
    if (!(*os->top)) exit(0); // exit at end of output
}
void push_os_file(char *filename){
    FILE *f = fopen(filename, "w");
    if (!f) error_file(filename);
    else push_os(f);
}
void push_os_fd(int fd){
    FILE *f = fdopen(fd, "w");
    if (!f) error_fd(fd);
    else push_os(f);
}
void output_printf(char *fmt, ...){
    va_list ap;
    va_start(ap, fmt);
    vfprintf(*os->top, fmt, ap);
    va_end(ap);
}

/* * * * * * * * * * * * * *
*          INPUT           *
* * * * * * * * * * * * * */

int eq(char *w){return !strcmp(word, w);}

void drop_is(){
    fclose (pop_is());
    if (!(*is->top)) exit(0); // exit at end of input
}
void push_is_file(char *filename){
    FILE *f = fopen(filename, "r");
    if (!f) error_file(filename);
    else push_is(f);
}
void push_is_fd(int fd){
    FILE *f = fdopen(fd, "r");
    if (!f) error_fd(fd);
    else push_is(f);
}

char input_getc(void){
    char c;
    while ((c = fgetc(*is->top)) == EOF) drop_is();
    return c;
}

void nextword(){
    char *pc = word;
  again:
    pc[0] = ' ';

    // skip whitespace
    do {pc[0] = input_getc();} while (isspace(pc[0]));
    
    // skip unix style comments
    if (pc[0] == '#'){
	do {pc[0] = input_getc();} while(pc[0] != '\n');
	goto again;
    }

    do {*++pc = input_getc();} while (!isspace(pc[0]));
    pc[0] = 0;

    // skip forth style comments
    while (eq("(")){
	do nextword();
	while (!eq(")"));
	nextword();
    }


}



/* * * * * * * * * * * * * *
*       DICTIONARY         *
* * * * * * * * * * * * * */


int dict_free(dictionary_t *dict){
    free(dict->item);
    free(dict);
}

void dict_add(dictionary_t *dict, char *name, void *address)
{
    item_t *i = dict->item_free++;
    i->address = address;
    i->name = malloc(strlen(name)+1);
    strcpy(i->name, name);
    //fprintf(stderr, "0x%08x %s\n", address, name);
}
int dict_rlookup(dictionary_t *dict, void *address, char **name)
{
   item_t *i = dict->item_free;
    while((--i) >= dict->item){
	if (address != i->address) continue;
	*name = i->name;
	return TRUE;
    }
    *name = 0;
    return FALSE; 
}
int dict_lookup(dictionary_t *dict, char *name, void **address)
{
    item_t *i = dict->item_free;
    while((--i) >= dict->item){
	if (strcmp(i->name, name)) continue;
	*address = i->address;
	if (i->address == 0) fprintf(stderr, "WARNING: NULL address\n");
	return TRUE;
    }
    *address = 0;
    return FALSE;
}

int dict_words(dictionary_t *dict)
{
    item_t *i = dict->item;
    while (i<dict->item_free){output_printf("%s ", i->name); i++;}
    output_printf("\n");
}

int word_lookup(int d, char *name, void **address){
    return dict_lookup(dict[d], name, address);
}

void word_words(void);
void word_host(void);
void word_target(void);
void word_file(void);

dictionary_t *dict_new(int item_size)
{
    dictionary_t *d = calloc(sizeof(*d),1);
    d->item = malloc(sizeof(item_t) * item_size);
    d->item_size = item_size;
    d->item_free = d->item;
    dict_add(d, "words",  word_words);
    dict_add(d, "host",   word_host);
    dict_add(d, "target", word_target);
    dict_add(d, "file",    word_file);
    return d;
}

// parse a word and compile to dict
int dict_compile_word(dictionary_t *dict){

    // lookup
    if (dict_lookup(dict, word, (void **)host_free)) {
	host_free++;
	return TRUE;
    }
    else {
	return FALSE;
    }
}


    

/* * * * * * * * * * * * * * *
*     COMPILER PRIMITIVES    *
* * * * * * * * * * * * * * */

void word_words(void) {dict_words(dict[dict_current]);}
void word_host(void)  {dict_current = HOST;}
void word_target(void){dict_current = TARGET;}


// a word is composite when it is in one of the dicts
int word_composite(prim_t *pp){
    return ((pp >= host_code) && (pp < host_free));
}

void word_lit(void){
    push_ds((int)(*host_ip++));
}
void word_print_ds(void) {stack_print("%d ", ds);}
void word_print_rs(void) {stack_print("0x%08x ", rs);}

// math
void word_add(void) {int i = pop_ds(); push_ds(pop_ds() + i);}
void word_sub(void) {int i = pop_ds(); push_ds(pop_ds() - i);}

void word_mul(void) {int i = pop_ds(); push_ds(pop_ds() * i);}
void word_div(void) {int i = pop_ds(); push_ds(pop_ds() / i);}
void word_mod(void) {int i = pop_ds(); push_ds(pop_ds() % i);}

void word_and(void) {int i = pop_ds(); push_ds(pop_ds() & i);}
void word_or (void) {int i = pop_ds(); push_ds(pop_ds() | i);}
void word_xor(void) {int i = pop_ds(); push_ds(pop_ds() ^ i);}
void word_not(void) {push_ds(pop_ds() ^ -1);}
void word_neg(void) {push_ds(-pop_ds());}

void word_shl(void) {int i = pop_ds(); push_ds(pop_ds() << i);}
void word_shr(void) {int i = pop_ds(); push_ds(pop_ds() >> i);}

#define COMP(op) {int b = pop_ds(); int a = pop_ds(); push_ds((a op b) ? -1 : 0);}
void word_lt(void) COMP(<)
void word_gt(void) COMP(>)
void word_eq(void) COMP(==)
void word_lte(void) COMP(<=)
void word_gte(void) COMP(>=)

// stack
void word_dup(void)  {push_ds(top_ds());}
void word_drop(void) {pop_ds();}
void word_swap(void) {
    int i = pop_ds();
    int j = pop_ds();
    push_ds(i);
    push_ds(j);
}
void word_over(void) {int *ids = (int *)ds->top; push_ds(ids[1]);}


// control & return stack
void word_trs(void) {push_ds((int)top_rs());}
void word_ret(void) {host_ip = pop_rs();}
void word_tor(void) {push_rs((prim_t*)pop_ds());}
void word_fromr(void) {push_ds((int)pop_rs());}
void word_here(void) {push_rs(host_ip-1);}

void word_for(void) {
    push_rs(host_ip); 
    word_tor();
}
void word_next(void) {
    int     *irs = (int *)(rs->top);
    prim_t **prs = (prim_t **)(rs->top);

    if (--(irs[0])) {
	host_ip = prs[1]; // continue looping
    }
    else {
	pop_rs(); // clean up return stack
	pop_rs();
    }
}

void word_skipz(void) {if (!pop_ds()) host_ip++;}

// bit of a juggle in our thread model
void word_bnz(void) {
    if (pop_ds()) {
	prim_t p = *host_ip;  // get next instruction
	prim_t *pp = (prim_t *)p;

	if (word_composite(pp)) host_ip = pp; // jump
	else {
	    p();        // execute primitive
	    word_ret(); // perform return
	}
    }
    else {
	host_ip++;
    }
}

void word_print(void) {output_printf("%d\n", pop_ds());}
void word_input(void) {nextword();}
void word_file(void) {nextword(); push_is_file(word);}


/* * * * * * * * * * * * * * * * * * * * * * * * *
*   COMPILATION AND EXECUTION FOR NATIVE FORTH   *
* * * * * * * * * * * * * * * * * * * * * * * * */

// parse the current word as a literal + compile
// if successful
int word_compile_lit(void)
{
    char *c;
    int _lit;

    _lit = strtol(word,&c,0);
    if (!c[0]) {
	// compile number
	host_free[0] = (prim_t)word_lit;
	host_free[1] = (prim_t)_lit;
	host_free += 2;
	return TRUE;
    }
    // not a valid number -> ERROR
    return FALSE;
    
    
}


/// execute a word (prim / composite)
void word_execute(prim_t p){

    // interprete p as prim_t* for composite (threaded) words
    prim_t *pp = (prim_t *)p; 

    // primitive -> execute
    if (!word_composite(pp)){p();}

    // composite -> interprete
    else {
	void **rsp = rs->top;     // record stack frame
	push_rs(host_ip);         // save previous next
	host_ip = pp;             // jump to threaded code
	while (rs->top != rsp){   // execute until we get to our stack frame
	    prim_t i = *host_ip++;   
	    word_execute(i);
	}
    }
}

// new dictionary entry (parsing word)
void word_newword(){
    nextword();
    dict_add(dict[dict_current], word, host_free);
}    

// do variable
void word_var(void){
    int address = host_ip - host_code;
    push_ds(address);
    host_ip = pop_rs();
}

// allocate n cells
void word_allot() {host_free += pop_ds();}

// create memory region (word indexed)
void word_create(){
    int address;
    word_newword();
    *host_free++ = (prim_t)word_var;
}

// compile word on stack
void word_spit() {*host_free++ = (prim_t)(pop_ds());}
void word_fetch() {push_ds((int)host_code[pop_ds() & (HOST_SIZE-1)]);}
void word_store() {int addr = pop_ds(); host_code[addr & (HOST_SIZE-1)] = (prim_t)pop_ds();}




// compile a new host word
void word_docolon(){

    word_newword();
    do {
	nextword();
	/* lookup and compile */
	if (!dict_compile_word(dict[HOST])){
	    if (!word_compile_lit()){
		error_undef(word);
		return;
	    }
	}
    } while (!eq(";"));
}


// hex files
typedef struct
{
    int length;
    int address;
    int type;    
    char *data;
} hex_t;

void output_char(char c)
{
    char tmp[3];
    sprintf(tmp, "%02x", (c&0xff));
    tmp[0] = toupper(tmp[0]);
    tmp[1] = toupper(tmp[1]);
    output_printf(tmp);
}
void output_start(void){output_printf(":");}


void output_end(void){output_printf(HEX_NEWLINE);}


// dump a hex record
#define OUT(x) {c = (x); checksum += c; output_char(c);}
void hex_output(hex_t *h){
    char c, checksum = 0;
    int i;
    output_start();
    OUT(h->length);
    OUT(h->address >> 8);
    OUT(h->address & 0xff);
    OUT(h->type);
    for (i=0; i<h->length; i++) OUT(h->data[i]);
    OUT(-checksum);
    output_end();
}
void hex_init(hex_t *h){memset(h, 0, sizeof(*h));}

// dump address in extended format
unsigned char byte(unsigned int i, int b){
    return (i >> (b*8)) & 0xff;
}


// set high 16bit of address
void hex_address_high(unsigned int address){
    hex_t h; 
    unsigned char a[2];
    hex_init(&h); h.length = 2; h.data = a;
    a[0] = byte(address, 3);
    a[1] = byte(address, 2);
    h.type = 4; 
    hex_output(&h);
}

void hex_end(void){
    hex_t h;
    hex_init(&h); 
    h.type = 1; 
    hex_output(&h);
}

void hex_data(unsigned int address, char *data, int length){
    hex_t h;
    hex_init(&h);
    h.length = length;
    h.address = address;
    h.data = data;
    hex_output(&h);
}

// dump entire target memory
void word_dump(void)
{
    unsigned char *ds = (char *)target_code;
    unsigned char *dx = (char *)target_free;
    unsigned char *d = ds;
    
    int blocksize = 16;
    int i = 0;
    hex_address_high(0x0);
    while (d  < dx) {
	hex_data(d-ds, d, blocksize);
	d += blocksize;
	i++;
    }
    hex_end();
}

void word_dumpfile(void){
    nextword();
    push_os_file(word);
    word_dump();
    drop_os();
}

void word_trace(void){
    trace = pop_ds();
}

/* * * * * * * * * *
*   TARGET MACROS  *
* * * * * * * * * */


// dump one word of code
void macro_spit(void){
    *target_free++ = pop_ds();
    //output_printf("COMPILED: 0x%04x\n", target_free[-1] & 0xffff);
}


// current compile point for target
// compiler addresses words (16bit)
// get free pointer
void macro_getpoint(void){push_ds((int)(2 * (target_free - target_code)));}
void macro_setpoint(void){target_free = &target_code[(pop_ds()/2) & (TARGET_SIZE - 1)];}
void macro_fetch(void){push_ds(target_code[(pop_ds()/2) & (TARGET_SIZE - 1)]);}
void macro_store(void){int addr = (pop_ds()/2) & (TARGET_SIZE-1); target_code[addr] = pop_ds();}

void macro_tlookup(void){
    void *i = 0;
    void *bi = target_code;
    nextword();
    dict_lookup(dict[TARGET], word, &i);
    push_ds((int)(i - bi));
}

/* * * * * * * * * * * * *
*   TARGET DICT & DEF    *
* * * * * * * * * * * * */

void word_1(int i, char *name){
    prim_t p;
    push_ds(i);
    if (!dict_lookup(dict[HOST], name, (void **)(&p))) error_undef(name);
    else word_execute(p);
}

void target_docolon()
{
    nextword();
    dict_add(dict[TARGET], word, (void **)target_free);

    do {
	prim_t p;
	char *end;
	unsigned short *addr;

	nextword();
	end = word + strlen(word);

	// check if word is a macro
	end[0] = ',';
	end[1] = 0;
	if (dict_lookup(dict[HOST], word, (void **)(&p))){
	    end[0] = 0;
	    word_execute(p);
	}
	else{
	    // check if it is a compiled word
	    end[0] = 0;
	    if (dict_lookup(dict[TARGET], word, (void **)(&addr))){
		int i = (addr - target_code) * 2;
		word_1(i, "call,");
	    }
	    else {
		// check if it is a number
		int i = strtoll(word, &end, 0);
		if (!end[0]) {
		    word_1(i, "lit,");
		}
		else {
		    // undefined
		    error_undef(word);
		    return;
		}
	    }
	}

    } while (!eq(";"));
}
	

// parse input in host/macro dictionary 
void target_escape(void){
    while (1){
	nextword();
	if (eq("]")) break;
	else {
	}
    }
}
	


/* * * * * * * * *
*   SETUP CODE   *
* * * * * * * * */



void setup(void)
{
    dictionary_t *d;
    
    /* setup stacks */
    ds = stack_new(32);
    rs = stack_new(32);
    os = stack_new(32);
    is = stack_new(32);


    /* setup compiler dictionary */
    d = dict[HOST] = dict_new(16*1024);
    dict_add(d, "create", word_create);  // create a memory referencing word
    
    dict_add(d, ":",      word_docolon); // define a compiler word
    dict_add(d, ";;",     word_ret);     // nonelectric ret
    dict_add(d, ";",      word_ret);     // 'electric' ret: will exit compile mode

    dict_add(d, ".S",     word_print_ds);
    dict_add(d, ".R",     word_print_rs);
    dict_add(d, "input",  word_input);
    dict_add(d, "dump",   word_dump);
    dict_add(d, "dumpfile",  word_dumpfile);
    dict_add(d, ".",         word_print);

    dict_add(d, "r",     word_trs);
    dict_add(d, ">r",    word_tor);
    dict_add(d, "r>",    word_fromr);
    dict_add(d, "here",  word_here);
    dict_add(d, "for",   word_for);
    dict_add(d, "next",  word_next);
    dict_add(d, "?",     word_skipz);
    dict_add(d, "?;",    word_bnz);

    dict_add(d, "dup",   word_dup);
    dict_add(d, "drop",  word_drop);
    dict_add(d, "swap",  word_swap);
    dict_add(d, "over",  word_over);

    dict_add(d, "+",     word_add);
    dict_add(d, "-",     word_sub);
    dict_add(d, "*",     word_mul);
    dict_add(d, "/",     word_div);
    dict_add(d, "mod",   word_mod);

    dict_add(d, "not",   word_not);
    dict_add(d, "neg",   word_neg);

    dict_add(d, ">>",   word_shr);
    dict_add(d, "<<",   word_shl);

    dict_add(d, "and",   word_and);
    dict_add(d, "or",    word_or);
    dict_add(d, "xor",   word_xor);

    dict_add(d, "<",     word_lt);
    dict_add(d, ">",     word_gt);
    dict_add(d, "=",     word_eq);
    dict_add(d, "<=",     word_lte);
    dict_add(d, ">=",     word_gte);

    dict_add(d, "!",     word_store);  // store to host mem
    dict_add(d, "@",     word_fetch);  // fetch from host mem

    dict_add(d, ",,",     word_spit);  // compile code/data to point in host
    dict_add(d, "allot", word_allot); // allocate n cells


    dict_add(d, ",",      macro_spit);    // compile code to point in target

    dict_add(d, "point",     macro_getpoint); // get/set target compile point
    dict_add(d, "!point",    macro_setpoint);

    dict_add(d, "p@",        macro_fetch);   // fetch/store from/to target program memory
    dict_add(d, "p!",        macro_store);

    dict_add(d, "tlookup",   macro_tlookup); // lookup target symbol

    dict_add(d, "trace",     word_trace); // lookup target symbol



    d = dict[TARGET] = dict_new(16*1024);
    dict_add(d, ":", target_docolon);

}

/* * * * * * * * * * * *
*  OUTER INTERPRETER   *
* * * * * * * * * * * */

int inputs(int argc, char **argv){
    push_is(0);                  // fallback input = zero (terminate program)
    if (argc == 0){
	push_is_fd(0);               // main input = stdin
    }
    while (argc--){
	push_is_file(argv[argc]);
    }
}



int main(int argc, char **argv)
{
    setup();                     // initialize
    dict_current = HOST;         // current dictionary is compiler words

    push_os(0);                  // fallback output = zero (terminate program)
    push_os_fd(1);               // main output = stdout

    inputs (argc-1, argv+1);

    while(1){

	nextword();

	prim_t p;
	char *c;
	int num;

	if (!(word_lookup(dict_current, word, (void **)(&p)))){
	    
	    // not found, try number
	    num = strtoll(word, &c, 0);
	    if (c[0]){
		error_undef(word);
	    }
            // load literal on data stack
	    else {
		push_ds(num); 
	    }
	}
	else {
	    // execute
	    word_execute(p);
	}

    }
    return 0;
}



