
( BADNOP dev env for PICmicro 18fxx2 )
( ********** 18f452 FORTH MACROS  ************ )
(

What you find here is the code generator for the 18fxx2 architecture.
All words ending with a ',' character are by convention words that generate
code at the current assembly point. They are invoked whenever a target
word's compilation encounters the name without the ',' character.

The compiler's data stack can be used to store syntactic data about
the word being compiled, like for instance is used in the implementation
of the for..next and if..then constructs.

Note that target's ':' word cannot load literals on the compiler stack.

)



( symbolic constants )
file 18f452.f


( INSTRUCTION SET )

( 18fxx2 opcodes )

: andwf  0x1400 ;
: addwf  0x2400 ;
: subwf  0x5C00 ;
: decf   0x0400 ;
: xorwf  0x1800 ;
: iorwf  0x1000 ;

: movwf  0x6E00 ;
: movf   0x5000 ;
: movlw  0x0E00 ;
: movff  0xC000 ;

: addlw  0x0F00 ;
: sublw  0x0800 ;
: andlw  0x0B00 ;
: xorlw  0x0A00 ;
: iorlw  0x0900 ;


: push   0x0005 ;
: pop    0x0006 ;

: bz     0xE000 ;
: bov    0xE400 ;
: bnov   0xE500 ;
: bnn    0xE700 ;
: bnc    0xE300 ;
: bn     0xE600 ;
: bc     0xE200 ;
: bnz    0xE100 ;
: bra    0xD000 ;
: rcall  0xD800 ;

: nop    0x0000 ;
: nopf   0xf000 ;   ( nop used as extension word for 2 word ins )
: return 0x0012 ;
: retfie 0x0010 ;
: call   0xEC00 ;
: goto   0xEF00 ;

: lfsr   0xEE00 ;


( instruction opcode modifiers : inclusive or masks )
: w/   ;                     ( destination wreg DEFAULT )
: reg/      0x0200 or ;      ( destination freg )
: access/ ;                  ( access bank DEFAULT )
: banked/   0x0100 or ;      ( bsr bank )
: op/       0xff and or ;    ( add operand to instruction )
: sop/      swap op/ ;

( forth opcodes  )
: op-drop movf  POSTDEC0 op/ w/ ;
: op-dup  movwf PREINC0  op/ reg/ ;

( peephole optimization state variables )
var lastdrop
var lastlit
var lastcall

: peep-off        ( turn off peephole optimization
                    this needs to be called at jump targets )
    0 lastdrop !
    0 lastlit !
    0 lastcall ! ;




( MACROS - map forth to assembler )

( utilities )
: undo    ( undo compilation of last instruction )
    point 2 - !point ;
: movff,  ( src dst -- )
    swap 0xfff and movff or ,
    0xfff and nopf or , ;
: lfsr,   ( value reg --  )
    >r dup
    8 >> 0x0f and lfsr r> 4 << or or , ( compile first word )
    0xff and nopf or , ;               ( compile second word )

( nop messages )
: bad,     0xfbad , ;



( stack manips )
: silentdrop,    POSTDEC0 WREG movff, ;   ( does not affect flags )
: test,          andlw 0xff or , ;        ( test Z and N flags from word on tos )
: >flags,        test, silentdrop, ;
    
    
( drop/save and opti )
: dup,    op-dup  , ;
: drop,
    point lastdrop !   ( record last drop )
    op-drop , ;
: got-drop?   lastdrop @ point 2 - = ;        ( last instruction a drop? )
: save,
    got-drop? not ?; dup,                     ( if prev instruction not a drop, compile dup )
    undo 0 lastdrop ! ;                       ( else undo last drop )


( jump/call/ret and opti )
: got-call? ( check if last instruction was a call )
    lastcall @ point 4 - = 
    point 4 - p@ 0xf000 and call = and ;
: tailcall, ( turn last call into a tail call )
    point 4 - p@ 
    0xfff and goto or
    point 4 - p!
    0 lastcall ! ;
: ;,
    got-call? ?; tailcall, return , ; ( compile a tail call or a return )
: ;;, ;, ;  ( fog )
: blong,    ( operands opcode -- ) ( compile a 2word long branch instruction )
    >r 2 / dup                     ( save opcode and convert to word addressing )
    0xff and r> or ,
    8 >> 0xff and nopf or , ;
: call,   point lastcall ! call blong, ;
: jump,   goto blong, ;



( literal and optimization
there are only 2 possible code transformations:
/dup lit/ -> //             undo the lit completely
/lit/     -> /drop/         undo an optimized lit
                            the drop can be undone by a following save, )
: lit,
    save,
    point lastlit !                ( save point for opti )
    movlw sop/ ,  ;
: lit2,                            ( , two byte literal )
    dup lit, 8 >> lit, ;
: get-lit                          ( get last literal value )
    lastlit @ p@ 0xff and ;
: got-lit?                         ( last instruction a literal? )
    point 2 - lastlit @ = ;
: undo-lit
    0 lastlit !                               ( disable further literal opti )
    undo                                      ( undo movlw ) 
    point 2 - p@ op-dup = ?; undo             ( "dup lit" -> "" )
    drop, ;                                   ( "lit" -> "drop" )
: not-lit?     got-lit? not ;
: re-lit       get-lit undo-lit ;             ( get literal and undo compilation )



( load/store,
what do we want to optimize? load-store sequences 1 @ 2 ! 3 @ 4 ! ...
should be a single dup, followed by load/store, followed by drop
this sequence calls the following macros: lit, @, lit, !, ... 
in general, this is nothing more than literal opti + drop/save
elimination. )

( set / store )
: regset,      movwf sop/ , ;
: set,         not-lit? ?; bad, re-lit regset, ;
: reg!,        regset, drop, ;
: !,           set, drop, ;

( get / fetch )
: regget,      movf sop/ , ;
: get,         not-lit? ?; bad, re-lit regget, ;
: reg@,        save, regget,  ;
: @,           not-lit? ?; bad, re-lit save, regget, ;



( doubles, not present in forth only as a literal macro )
: reg!!,       dup 1 + reg!, reg!, ;          ( store low high to address )
: reg@@,       dup reg@, reg@, ;              ( fetch low high from address )


( TOS words )
: neg,      sublw , ;
: not,      xorlw 0xff op/ , ;

( 2 argument words  )
: 2op,      POSTDEC0 op/ reg/ , ;      ( compile stack 2op instruction )

: real+,    addwf 2op, ;
: real-,    subwf 2op, ;   
: realxor,  xorwf 2op, ;
: realand,  andwf 2op, ;
: realor,   iorwf 2op, ;

( arithmic operating on TOS and immediate )
: lit+,     addlw sop/ , ;               ( compile literal add )
: lit-,     neg lit+, ;                  ( literal sub )
: litxor,   xorlw sop/ , ;
: litand,   andlw sop/ , ;
: litor,    iorlw sop/ , ;


( literal folding )
: +,        not-lit?  ?; real+,    re-lit lit+, ;
: -,        not-lit?  ?; real-,    re-lit lit-, ;
: or,       not-lit?  ?; realor,   re-lit litor, ;
: xor,      not-lit?  ?; realxor,  re-lit litxor, ;
: and,      not-lit?  ?; realand,  re-lit litand, ;



( return stack )
: TOS--,  decf TOSL op/ reg/ , ;           ( decrement low byte of word on RS )
: here,   push , TOS--, TOS--, ;           ( push location )
: >r,     push , TOSL reg!, ;              ( move byte to/from return stack )
: r>,     TOSL reg@, pop , ;
: >>r,    push , TOSH reg!, TOSL reg!, ;   ( move word to/from return stack )
: r>>,    TOSL reg@, TOSH reg@, pop , ;
: rdrop,  pop , ;
: rl,     TOSL reg@, ;                     ( get low/high byte from RS )
: rh,     TOSH reg@, ;



( conditional blocks use the machine flags )

: encode-srcdst   ( to from -- jmpoperand )
    -             ( distance )
    2 -           ( - 2 to compensate for PC inc. )
    2 / ;         ( instruction uses word addressing )
: offset-to    point encode-srcdst  ;              ( compute branch offset point->address )
: offset-from  point swap encode-srcdst ;          (                       address->point )
: rjump,       offset-to 0x7ff and bra or , ;      ( short relative jump )
: rcall,       offset-to  0x7ff and rcall or , ;   ( short relative call )
: op-cjump
    ( address opcode -- packedopcode ) ( create conditional jump opcode )
    >r offset-to 0xff and r> or ;
: patch-cjump, ( addr )
    dup offset-from 0xff and >r  ( compute relative distance ) 
    dup p@                       ( get opcode )
    0xff00 and r> or             ( strip address and insert new )
    swap p! ;                    ( store instruction )
: for,
    >r, point        ( move count to r, save point on compiler stack )
    peep-off ;       ( disable peephole opti )
: next,              ( decrement value in TOSL and branch back if nonzero )
    decf TOSL op/ reg/ ,
    bnz op-cjump ,
    pop , ;
: ifnz,  point bz , ;
: ifz,   point bnz , ;
: ifc,   point bnc , ;
: ifnc,  point bc , ;
: ifn,   point bnn , ;
: ifnn,  point bn , ;
: ifov,  point bnov , ;
: ifnov, point bov , ;
: skip,  point bra , ;
: if,    >flags, ifnz, ;   ( proper if uses tos. use flag ifs if possible. they are faster )
: then,                    ( can be used to patch any conditional reljump )
    patch-cjump,           ( it can patch forward bra too )
    peep-off ;
: else,                    ( patch first cond jump and set new uncond )
    skip, >r patch-cjump, r> ;
: nop, nop , ;





( interrupt handlers )

: reset, ( startword -- )
    point >r ( save point )

    ( compile relative jump to start routine )
    0x0 !point      ( compile at 0x0 )
    0x40 rjump,     ( jump to 0x40 )

    ( init routine: setup forth environment )
    0x40 !point         ( compile at 0x40 )
    movlw ,             ( clear TOS )
    0x0 0 lfsr,         ( init data stack ptr )
    movwf STKPTR op/ ,  ( init return stack ptr )
    0x100 2 lfsr,       ( init IP for threaded code )
    rjump,              ( jump to start word )
           
    r> !point ; ( restore point )

: ihigh, ( intword -- )
    point >r

    0x8 !point          ( move point to iv )
    call,               ( call handling word )
    retfie 1 or ,       ( fast return )
    
    r> !point ;

: ilow, ( intword -- ) 
    point >r
    0x18 !point

    STATUS reg@,
    BSR reg@,
    call,
    BSR reg!,
    STATUS reg!,
    retfie ,

    r> !point ;

: interpreter, ( interpreter -- )
    point >r

    0xF0 !point         ( install trampoline at 0x00F0 )
    TOSL reg@, 2 lit-,  ( get low byte of original address )
    1 lit,              ( load high byte )
    pop ,               ( pop return stack )
    rjump,              ( jump to interpreter )

    0x100 !point        ( compile stubs )
    0x80 for 0xF0 rcall, next
    
    r> !point ;

( some macros used in the interpreter )
: ip>r,
    push ,
    FSR2L reg@, TOSL reg!,
    FSR2H reg@, TOSH reg!, ;
: r>ip,
    TOSL reg@, FSR2L reg!,
    TOSH reg@, FSR2H reg!, 
    pop , ;
: ip++>r,
    push ,
    POSTINC2 reg@, TOSH reg!,
    POSTINC2 reg@, TOSL reg!, ;
: ip!,
    FSR2H reg!, FSR2 ;
    
    

file 18f452_mac.f   ( load symbolic name macros )
0x200 !point        ( start compiling forth code at 0x200 ) 




( target kernel code )
target


(
small script interpreter.
should use the return stack in a way that we can have
caller thread thread thread ..
we just use an expensive next

TODO: fix this

)


: NEXT          ( execute next primitive )

    ( instruction layout
    00: immediate bytes
    01: threaded script word
    02: call threaded forth word
    ff: exit )

    
    
    
    here        ( primitive words jump to next through return )
    ip++>r ;    ( get next instruction on return stack )
    
: ENTER         ( enter high level word )
    ip>r ip! NEXT ;

: LEAVE
    rh                  ( get high byte of return address )
    1 - ifz r>ip then   ( if primitive, pop ip )
    
    
    rdrop       ( remove next's address from return stack )
    r>ip ;      ( restore )
    
    

: interprete
    ip>r     ( save previous ip )
    ip!      ( set new )

    here ip++>r ;;  ( loopget next )

    r>ip     
;
    

host
